/**
 * Copyright (C) 2009 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include <glib/gstdio.h>

#include "wk-key-file.h"

#include "utils.h"
#include "main.h"
#include "settings.h"

void
wk_settings_set_defaults (WkSettings **settings)
{
    WkSettings *lsettings;

    lsettings = g_new (WkSettings, 1);

    lsettings->can_quit = PREF_CAN_QUIT;
    lsettings->can_end_session = PREF_CAN_END_SESSION;
    lsettings->fullscreen = PREF_FULLSCREEN;
    lsettings->with_timer = PREF_WITH_TIMER;
    lsettings->with_clear = PREF_WITH_CLEAR;
    lsettings->default_url = g_strdup (PREF_DEFAULT_URL);
    lsettings->max_history = PREF_MAX_HISTORY;
    lsettings->max_windows = PREF_MAX_WINDOWS;

    lsettings->list_mode = LIST_MODE_NONE;
    lsettings->list = NULL;

    lsettings->prestart = NULL;
    lsettings->start_with_session = PREF_START_WITH_SESSION;
    lsettings->start_with_time = PREF_START_WITH_TIME;

    lsettings->post_session_uri = NULL;
    lsettings->post_session_uri_time = PREF_POST_SESSION_URI_TIME;

    lsettings->slideshow_views = NULL;
    lsettings->slideshow_hide_pointer = PREF_SLIDESHOW_HIDE_POINTER;

    lsettings->mime_associations = NULL;

    lsettings->admin_passwd = NULL;

    lsettings->tmpdir_name = g_strdup_printf ("webkiosk-%d", getuid ());

    /* we assume that settings is not allocated yet */
    *settings = lsettings;
}

void
wk_settings_get (WkSettings **settings, WkKeyFile *kf)
{
    WkSettings *lsettings = NULL;
    gchar *tmp = NULL;
    gchar **types = NULL;
    gboolean list_use_regex;
    gchar *contents = NULL;
    gchar *uri = NULL;

    lsettings = g_new (WkSettings, 1);

    /* set a few things */
    lsettings->list = NULL;
    lsettings->slideshow_views = NULL;
    lsettings->post_session_uri = NULL;

    lsettings->can_quit = wk_key_file_get_boolean (
        kf, "main", "can_quit", PREF_CAN_QUIT);
    lsettings->can_end_session = wk_key_file_get_boolean (
        kf, "main", "can_end_session", PREF_CAN_END_SESSION);
    lsettings->fullscreen = wk_key_file_get_boolean (
        kf, "main", "fullscreen", PREF_FULLSCREEN);
    lsettings->with_timer = wk_key_file_get_boolean (
        kf, "main", "with_timer", PREF_WITH_TIMER);
    lsettings->with_clear = wk_key_file_get_boolean (
        kf, "main", "with_clear", PREF_WITH_CLEAR);
    lsettings->max_history = wk_key_file_get_int (
        kf, "main", "max_history", PREF_MAX_HISTORY);
    lsettings->max_windows = wk_key_file_get_int (
        kf, "main", "max_windows", PREF_MAX_WINDOWS);

    lsettings->default_url = wk_key_file_get_string (
        kf, "main", "default_url", PREF_DEFAULT_URL);

    lsettings->post_session_uri_time = wk_key_file_get_int (
        kf, "main", "post_session_uri_time", PREF_POST_SESSION_URI_TIME);
    lsettings->post_session_uri = wk_key_file_get_string (
        kf, "main", "post_session_uri", "");
    if (g_strcmp0 (lsettings->post_session_uri, "") == 0)
    {
        g_free (lsettings->post_session_uri);
        lsettings->post_session_uri = NULL;
    }

    lsettings->prestart = wk_key_file_get_string (
        kf, "main", "prestart", NULL);
    if ((lsettings->prestart)
        && !g_file_test (lsettings->prestart, G_FILE_TEST_IS_EXECUTABLE))
    {
        g_free (lsettings->prestart);
        lsettings->prestart = NULL;
    }

    /* use start_with_session only if we don't have a valid prestart script */
    if (!lsettings->prestart)
    {
        lsettings->start_with_session = wk_key_file_get_boolean (
            kf, "main", "start_with_session", PREF_START_WITH_SESSION);
        if (lsettings->start_with_session && lsettings->with_timer)
            lsettings->start_with_time = wk_key_file_get_int (
                kf, "main", "start_with_time", PREF_START_WITH_TIME);
    }

    tmp = wk_key_file_get_string (kf, "main", "list_mode", NULL);
    if (g_strcmp0 (tmp, "whitelist") == 0)
        lsettings->list_mode = LIST_MODE_WHITELIST;
    else if (g_strcmp0 (tmp, "blacklist") == 0)
        lsettings->list_mode = LIST_MODE_BLACKLIST;
    else
        lsettings->list_mode = LIST_MODE_NONE;
    g_free (tmp);
    tmp = NULL;

    list_use_regex = wk_key_file_get_boolean (
        kf, "main", "list_use_regex", PREF_LIST_USE_REGEX);

    if (lsettings->list_mode != LIST_MODE_NONE)
    {
        gchar **lines = NULL;
        gchar **linep = NULL;

        if ((contents = wk_key_file_get_string_from_uri (
                            kf, "main", "list_uri", &uri, NULL)) != NULL)
        {
            lines = g_strsplit (contents, "\n", -1);
            for (linep = lines; *linep; linep++)
            {
                if (strlen (*linep) == 0)
                    continue;

                gchar *regex;
                regex = list_use_regex ? g_strdup (*linep)
                                       : wildcard_to_regex (*linep);
                if (regex)
                    lsettings->list = g_list_append (lsettings->list, regex);

                /* the pointer to *regex is held by the list, we don't free it */
            }
            g_free (contents);
            contents = NULL;
            g_strfreev (lines);
        }
        if (uri)
        {
            g_free (uri);
            uri = NULL;
        }
    }

    lsettings->slideshow_hide_pointer = wk_key_file_get_boolean (
        kf, "slideshow", "hide_pointer", PREF_SLIDESHOW_HIDE_POINTER);
    lsettings->slideshow_views = NULL;
    if ((contents = wk_key_file_get_string_from_uri (
                        kf, "slideshow", "pages", &uri, NULL)) != NULL)
    {
        gchar **lines = NULL;
        gchar **linep = NULL;
        gchar **spl = NULL;

        lines = g_strsplit (contents, "\n", -1);
        for (linep = lines; *linep; linep++)
        {
            if (strlen (*linep) == 0)
                continue;

            spl = g_strsplit (*linep, " ", 0);
            if (g_strv_length (spl) != 2)
            {
                g_strfreev (spl);
                spl = NULL;
                continue;
            }
            SlideshowItem *item = g_new (SlideshowItem, 1);
            item->uri = g_strdup (spl[0]);
            item->time = atoi (spl[1]);

            lsettings->slideshow_views = g_list_append (
                lsettings->slideshow_views, item);
            g_strfreev (spl);
            spl = NULL;
        }
        g_strfreev (lines);
        g_free (contents);
        contents = NULL;
    }

    lsettings->mime_associations = g_hash_table_new (g_str_hash, g_str_equal);
    types = wk_key_file_get_keys (kf, "mimetypes", NULL, NULL);
    if (G_LIKELY (types))
    {
        gchar **ptr;
        for (ptr = types; *ptr; ptr++)
        {
            /*
             * if no application is specified, we try to open the file through
             * xdg methods
             */
            g_hash_table_insert (lsettings->mime_associations,
                                 *ptr,
                                 wk_key_file_get_string (kf, "mimetypes",
                                                         *ptr, "xdg_open"));
        }
        /* free the array but not its content, held by the hash table */
        g_free (types);
        types = NULL;
    }

    lsettings->admin_passwd = wk_key_file_get_string (
        kf, "main", "admin_passwd", NULL);

    lsettings->tmpdir_name = g_strdup_printf ("webkiosk-%d", getuid ());

    g_object_unref (kf);

    /* we assume that settings is not allocated yet */
    *settings = lsettings;
}

void
wk_settings_load_from_uri (WkSettings **settings, const gchar *uri)
{
    WkKeyFile *kf;

    kf = wk_key_file_new ();

    if (wk_key_file_load_from_uri (kf, uri))
        wk_settings_get (settings, kf);
    else
        wk_settings_set_defaults (settings);
}

static void
slideshow_free_each (gpointer a, gpointer b)
{
    SlideshowItem *item = (SlideshowItem *) a;
    g_free ((gchar *)item->uri);
    g_free (item);
}

static void
list_free_each (gpointer a, gpointer b)
{
    g_free ((gchar *) a);
}

void
wk_settings_free (WkSettings *settings)
{
    g_free (settings->default_url);
    g_free (settings->prestart);
    g_free (settings->tmpdir_name);

    g_list_foreach (settings->list, list_free_each, NULL);
    g_list_free (settings->list);

    g_list_foreach (settings->slideshow_views, slideshow_free_each, NULL);
    g_list_free (settings->slideshow_views);

    g_hash_table_destroy (settings->mime_associations);
}




/* tests! */
static void
_settings_append_dir_to_list (const gchar *dirname, gint t, GList **l)
{
    GDir *dir;
    gchar *filename = NULL, *uri = NULL;
    const gchar *localname = NULL;
    SlideshowItem *item;

    if ((dir = g_dir_open (dirname, 0, NULL)) == NULL)
        return;

    while ((localname = g_dir_read_name (dir)) != NULL)
    {
        uri = g_strdup_printf ("file://%s/%s", dirname, localname);
        filename = uri+7;

        if (g_file_test (filename, G_FILE_TEST_IS_DIR))
        {
            _settings_append_dir_to_list (filename, t, l);
            g_free (uri);
        }
        else
        {
            item = g_new (SlideshowItem, 1);
            item->uri = uri;
            item->time = t;
            *l = g_list_append (*l, item);
        }
    }
    g_dir_close (dir);
}

static void
_settings_expand_slideshow_settings (SlideshowItem *item, GList **l)
{
    SlideshowItem *new;

    if (!g_file_test ((item->uri)+7, G_FILE_TEST_IS_DIR))
    {
        new = g_new (SlideshowItem, 1);
        new->uri = g_strdup (item->uri);
        new->time = item->time;
        *l = g_list_append (*l, new);
        return;
    }

    _settings_append_dir_to_list ((item->uri)+7, item->time, l);
}

GList *
wk_settings_expand_playlist (WkSettings *settings)
{
    GList *new_list = NULL;

    if (!settings->slideshow_views)
        return NULL;

    g_list_foreach (settings->slideshow_views,
                    (GFunc) _settings_expand_slideshow_settings,
                    &new_list);

    return new_list;
}

