/**
 * Copyright (C) 2009 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <glib/gstdio.h>

#include "settings.h"
#include "gtk-extensions.h"

#include "config-display.h"

#define DISPLAYURI SYSCONFDIR "/xdg/webkiosk/playlist"
#define DEFAULT_TIME 5

static GHashTable *hash;
static GtkWidget *treeview;
static GtkListStore *store;
static GtkWidget *add_btn, *rm_btn;

enum
{
    TARGET_OWN_WIDGET,
    TARGET_URI_LIST,
    TARGET_UTF8_STRING,
    TARGET_TEXT,
    TARGET_COMPOUND_TEXT,
    TARGET_STRING,
    TARGET_TEXT_PLAIN,
    TARGET_MOZ_URL,
};

static const GtkTargetEntry dnd_targets[] =
{
    { "local_target", GTK_TARGET_SAME_WIDGET, TARGET_OWN_WIDGET },
    { "text/uri-list", 0, TARGET_URI_LIST },
    { "text/x-moz-url", 0, TARGET_MOZ_URL },
    { "UTF8_STRING", 0, TARGET_UTF8_STRING },
    { "TEXT", 0, TARGET_TEXT },
    { "COMPOUND_TEXT", 0, TARGET_COMPOUND_TEXT },
    { "STRING", 0, TARGET_STRING },
    { "text/plain", 0, TARGET_TEXT_PLAIN },
};

ConfigItemDesc display_descs[] =
{
    { "slideshow", "pages", CONFIG_ITEM_TYPE_STRING, "" },
    { "slideshow", "hide_pointer", CONFIG_ITEM_TYPE_BOOLEAN, "true" },
    { NULL }
};

inline GHashTable *
config_display_get_hash (WkKeyFile *kf)
{
    hash = handled_settings_get_from_array (kf, display_descs);
    return hash;
}

static void
populate_store ()
{
    const gchar *listfile;
    gchar *contents = NULL;
    gchar **lines = NULL;
    gchar **linep = NULL;
    gchar **spl = NULL;
    ConfigItem *item;
    GtkTreeIter iter;

    item = g_hash_table_lookup (hash, "slideshow:pages");
    if (strcmp ((gchar *) wk_config_item_get_value (item), "") == 0)
    {
        listfile = DISPLAYURI;
        wk_config_item_set_value (item, (gpointer) g_strdup (listfile), TRUE);
    }
    else
        listfile = (gchar *) wk_config_item_get_value (item);

    if (! g_file_get_contents (listfile, &contents, NULL, NULL))
        return;

    lines = g_strsplit (contents, "\n", -1);
    for (linep = lines; *linep; linep++)
    {
        if (strlen (*linep) == 0)
            continue;

        spl = g_strsplit (*linep, " ", 0);
        if (g_strv_length (spl) != 2)
        {
            g_strfreev (spl);
            spl = NULL;
            continue;
        }

        gtk_list_store_append (store, &iter);
        gtk_list_store_set (store, &iter,
                            0, g_strdup (spl[0]),
                            1, atoi (spl[1]),
                            -1);

        g_strfreev (spl);
        spl = NULL;
    }
    g_strfreev (lines);
    g_free (contents);
}

static void
on_url_edited (GtkCellRendererText *renderer,
               gchar *path,
               gchar *new_text,
               gpointer data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;

    model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
    if (!gtk_tree_model_get_iter_from_string (model, &iter, path))
        return;

    if (strcmp (new_text, "") != 0)
        gtk_list_store_set (GTK_LIST_STORE (model), &iter, 0, new_text, -1);
}

static void
on_spin_edited (GtkCellRendererText *renderer,
                gchar *path,
                gchar *new_text,
                gpointer data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    int value;

    model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
    if (!gtk_tree_model_get_iter_from_string (model, &iter, path))
        return;

    if ((value = atoi (new_text)) > 0)
        gtk_list_store_set (GTK_LIST_STORE (model), &iter, 1, value, -1);
}

static void
on_treeview_changed (GtkTreeView *treeview, gpointer data)
{
    GtkTreeIter iter;
    GtkTreeModel *model = GTK_TREE_MODEL (store);

    if (gtk_tree_selection_get_selected (
            gtk_tree_view_get_selection (treeview), &model, &iter))
        gtk_widget_set_sensitive (rm_btn, TRUE);
    else
        gtk_widget_set_sensitive (rm_btn, FALSE);
}

static void
run_add_dialog (const gchar *init_uri, gint init_time)
{
    GtkWidget *d, *add_entry, *add_time;
    GtkWidget *table, *label;
    const gchar *tmp;
    gchar *uri = NULL;
    gint time;
    GtkTreeIter iter;

    d = gtk_dialog_new_with_buttons (_("Add an item"),
                                     NULL,
                                     GTK_DIALOG_MODAL | GTK_DIALOG_NO_SEPARATOR,
                                     GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                     GTK_STOCK_ADD, GTK_RESPONSE_OK,
                                     NULL);

    table = gtk_table_new (2, 3, FALSE);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox), table, TRUE, TRUE, 10);

    label = gtk_label_new (_("Type the URL or file path (image, video or folder) "
                             "that you want to add to the slideshow:"));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
    gtk_table_attach (GTK_TABLE (table), label,
                      0, 2, 0, 1, GTK_SHRINK, GTK_SHRINK, 0, 6);

    label = gtk_label_new (_("URL or path: "));

    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_table_attach (GTK_TABLE (table), label,
                      0, 1, 1, 2, GTK_SHRINK, GTK_SHRINK, 3, 0);

    label = gtk_label_new (_("Time (seconds): "));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_table_attach (GTK_TABLE (table), label,
                      0, 1, 2, 3, GTK_SHRINK, GTK_SHRINK, 3, 0);

    add_entry = gtk_entry_new ();
    if (init_uri)
        gtk_entry_set_text (GTK_ENTRY (add_entry), init_uri);
    gtk_table_attach (GTK_TABLE (table), add_entry,
                      1, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);

    add_time = gtk_spin_button_new_with_range (0, 3600, 1);
    gtk_spin_button_set_value (GTK_SPIN_BUTTON (add_time),
        init_time >= 0 ? init_time : DEFAULT_TIME);
    gtk_table_attach (GTK_TABLE (table), add_time,
                      1, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 0);

    gtk_widget_show_all (table);

    if (gtk_dialog_run (GTK_DIALOG (d)) != GTK_RESPONSE_OK)
        goto out;

    tmp = gtk_entry_get_text (GTK_ENTRY (add_entry));
    if (strlen (tmp) < 1)
        goto out;

    if (g_str_has_prefix (tmp, "/"))
        uri = g_strdup_printf ("file://%s", tmp);
    else
        uri = g_strdup (tmp);

    time = gtk_spin_button_get_value (GTK_SPIN_BUTTON (add_time));
    if (time <= 0)
        goto out;

    gtk_list_store_append (store, &iter);
    gtk_list_store_set (store, &iter, 0, uri, 1, time, -1);

out:
    gtk_widget_destroy (d);
    return;
}

static void
on_add_btn_clicked (GtkButton *btn, gpointer data)
{
    run_add_dialog (NULL, DEFAULT_TIME);
}

static void
on_treeview_drag_data_received (GtkWidget *treeview,
                                GdkDragContext *drag_context,
                                gint x, gint y,
                                GtkSelectionData *selection_data,
                                guint info, guint time,
                                gpointer data)
{
    GtkTreeIter src_iter, dest_iter;
    GtkTreePath *dest_path = NULL, *src_path = NULL;
    GtkTreeViewDropPosition pos;
    GtkTreeModel *model;
    gchar *text = NULL;
    gchar **uris = NULL;
    gchar **urip;

    model = GTK_TREE_MODEL (store);

    /* informations about source path */
    if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection(
        GTK_TREE_VIEW (treeview)), NULL, &src_iter))
    {
        src_path = gtk_tree_model_get_path(model, &src_iter);
    }
    else
        src_path = NULL;

    /* informations about destination path */
    gtk_tree_view_get_dest_row_at_pos (GTK_TREE_VIEW (treeview),
                                           x, y, &dest_path, &pos);
    if (dest_path)
        gtk_tree_model_get_iter(model, &dest_iter, dest_path);

    switch (info)
    {
        case TARGET_STRING:
        case TARGET_UTF8_STRING:
        case TARGET_COMPOUND_TEXT:
        case TARGET_TEXT:
            text = (gchar *) gtk_selection_data_get_text (selection_data);
            if (text == NULL)
                break;

            if (pos == GTK_TREE_VIEW_DROP_BEFORE
                || pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
                    gtk_list_store_insert_before (store, &dest_iter, NULL);
            else
                    gtk_list_store_insert_after (store, &dest_iter, NULL);

            gtk_list_store_set (GTK_LIST_STORE (store), &dest_iter,
                                0, text,
                                1, DEFAULT_TIME,
                                -1);
            break;

        case TARGET_TEXT_PLAIN:
            if (pos == GTK_TREE_VIEW_DROP_BEFORE
                || pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
                    gtk_list_store_insert_before (store, &dest_iter, NULL);
            else
                    gtk_list_store_insert_after (store, &dest_iter, NULL);
            gtk_list_store_set (GTK_LIST_STORE (store), &dest_iter,
                                0, g_strdup ((gchar *) selection_data->data),
                                1, DEFAULT_TIME,
                                -1);
            break;

        case TARGET_URI_LIST:
            uris = g_uri_list_extract_uris ((gchar *) selection_data->data);

            if (dest_path && !src_path)
                gtk_list_store_prepend (GTK_LIST_STORE (store), &dest_iter);
            else if (!dest_path && !src_path)
                gtk_list_store_append (GTK_LIST_STORE (store), &dest_iter);
            else
            {
                if (pos == GTK_TREE_VIEW_DROP_BEFORE
                    || pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
                {
                    gtk_list_store_insert_before (GTK_LIST_STORE (store), &dest_iter,
                                                  dest_path ? &dest_iter : NULL);
                }
                else
                    gtk_list_store_insert_after (GTK_LIST_STORE (store), &dest_iter,
                                                 dest_path ? &dest_iter : NULL);
            }
            gtk_list_store_set (GTK_LIST_STORE (store), &dest_iter,
                                0, *uris,
                                1, DEFAULT_TIME,
                                -1);

            /* set the others after the row we just inserted */
            for (urip = uris+1; *urip; urip++)
            {
                gtk_list_store_insert_after (GTK_LIST_STORE (store), &dest_iter, &dest_iter);
                gtk_list_store_set (GTK_LIST_STORE (store), &dest_iter,
                                    0, *urip,
                                    1, DEFAULT_TIME,
                                    -1);
            }
            g_strfreev (uris);
            break;

        case TARGET_OWN_WIDGET:
            if (!src_path || !dest_path)
                break;

            if (pos == GTK_TREE_VIEW_DROP_BEFORE
                || pos == GTK_TREE_VIEW_DROP_INTO_OR_BEFORE)
                    gtk_list_store_move_before (store, &src_iter, &dest_iter);
            else
                    gtk_list_store_move_after (store, &src_iter, &dest_iter);

            gtk_drag_finish (drag_context, TRUE, FALSE, time);

            break;
    }

    gtk_tree_path_free (src_path);
    gtk_tree_path_free (dest_path);
}

static void
on_rm_btn_clicked (GtkButton *btn, gpointer data)
{
    GtkTreeIter iter;
    GtkTreeModel *model = GTK_TREE_MODEL (store);

    if (!gtk_tree_selection_get_selected (
        gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)), &model, &iter))
        return;

    gtk_list_store_remove (store, &iter);

    on_treeview_changed (GTK_TREE_VIEW (treeview), NULL);
}

GtkWidget *
config_display_get_widget (GHashTable *hash)
{
    GtkWidget *mvbox; /* returned widget */
    GtkWidget *scw, *hbox, *vbox;
    GtkWidget *label;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    GtkAdjustment *adjustment;

    mvbox = gtk_vbox_new (FALSE, 2);

    gtk_box_pack_start (GTK_BOX (mvbox),
                        wk_gtk_check_button_new (_("Hide the mouse pointer"),
                            g_hash_table_lookup (hash, "slideshow:hide_pointer"),
                            FALSE),
                        FALSE, FALSE, 0);

    gtk_box_pack_start (GTK_BOX (mvbox), gtk_hseparator_new (), FALSE, FALSE, 0);

    hbox = gtk_hbox_new (FALSE, 2);

    vbox = gtk_vbox_new (FALSE, 2);
    gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 2);

    scw = gtk_scrolled_window_new (NULL, NULL);
    gtk_box_pack_start (GTK_BOX (vbox), scw, TRUE, TRUE, 0);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scw),
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC);

    label = gtk_label_new (_("Drag and drop items from the file manager to add "
                             "them to the list."));
    gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

    store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
    treeview = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
    gtk_container_add (GTK_CONTAINER (scw), treeview);
    gtk_tree_view_enable_model_drag_source (GTK_TREE_VIEW (treeview),
                                            GDK_BUTTON1_MASK,
                                            dnd_targets,
                                            G_N_ELEMENTS (dnd_targets),
                                            GDK_ACTION_MOVE);
    gtk_tree_view_enable_model_drag_dest (GTK_TREE_VIEW (treeview),
                                          dnd_targets,
                                          G_N_ELEMENTS (dnd_targets),
                                          GDK_ACTION_DEFAULT);

    renderer = gtk_cell_renderer_text_new ();
    g_object_set (renderer, "editable", TRUE, NULL);
    column = gtk_tree_view_column_new_with_attributes (
        _("URL or path"), renderer, "text", 0, NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
    g_signal_connect (G_OBJECT (renderer), "edited",
                      G_CALLBACK (on_url_edited), NULL);

    renderer = gtk_cell_renderer_spin_new ();
    adjustment = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 3600, 1, 10, 10));
    g_object_set (renderer, "editable", TRUE, NULL);
    g_object_set (renderer, "adjustment", adjustment, NULL);
    column = gtk_tree_view_column_new_with_attributes (
        _("Time"), renderer, "text", 1, NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
    g_signal_connect (G_OBJECT (renderer), "edited",
                      G_CALLBACK (on_spin_edited), NULL);

    populate_store ();

    vbox = gtk_vbox_new (FALSE, 2);
    gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 2);

    add_btn = gtk_button_new_from_stock (GTK_STOCK_ADD);
    rm_btn = gtk_button_new_from_stock (GTK_STOCK_DELETE);
    gtk_widget_set_sensitive (rm_btn, FALSE);

    gtk_box_pack_start (GTK_BOX (vbox), add_btn, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), rm_btn, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (mvbox), hbox, TRUE, TRUE, 2);

    g_signal_connect (G_OBJECT (treeview), "cursor-changed",
                      G_CALLBACK (on_treeview_changed), NULL);
    g_signal_connect (G_OBJECT (treeview), "drag-data-received",
                      G_CALLBACK (on_treeview_drag_data_received), NULL);
    g_signal_connect (G_OBJECT (add_btn), "clicked",
                      G_CALLBACK (on_add_btn_clicked), NULL);
    g_signal_connect (G_OBJECT (rm_btn), "clicked",
                      G_CALLBACK (on_rm_btn_clicked), NULL);

    return mvbox;
}

static gboolean
append_data (GtkTreeModel *model, GtkTreePath *path,
             GtkTreeIter *iter, gpointer data)
{
    gchar *line = NULL;
    gint time;
    gchar *tmp = NULL;
    gchar **datap = (gchar **) data;

    gtk_tree_model_get (model, iter, 0, &line, 1, &time, -1);

    if (*datap == NULL)
        *datap = g_strdup_printf ("%s %d", line, time);
    else
    {
        tmp = g_strdup_printf ("%s\n%s %d", *datap, line, time);
        g_free (*datap);
        *datap = tmp;
    }

    g_free (line);

    return FALSE;
}

inline void
config_display_save_config (WkKeyFile *kf, GHashTable *hash)
{
    ConfigItem *item;
    const gchar *filename;
    gchar *tmpname = NULL;
    gchar *data = NULL;

    handled_settings_set_from_hash (kf, hash);

    item = g_hash_table_lookup (hash, "slideshow:pages");
    filename = (const gchar *) wk_config_item_get_value (item);
    tmpname = g_strdup_printf ("%s.tmp", filename);

    gtk_tree_model_foreach (GTK_TREE_MODEL (store), append_data, &data);

    if (!data)
        data = g_strdup ("");

    if (!g_file_set_contents (tmpname, data, -1, NULL)
        || (g_rename (tmpname, filename) != 0))
    {
        gtk_dialog_run (GTK_DIALOG (gtk_message_dialog_new (
            NULL, 0, GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            _("Unable to save the listing file "
            "to\n%s"), filename)));
    }

    g_free (tmpname);
    g_free (data);
}
